/*
    Copyright (C) 1998, 1999  Michael Ravits

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "include/newplayer.h"
#include "include/screen.h"
#include "include/stuff.h"
#include "include/player_newplayer.h"
#include "include/playlist.h"


stuff::stuff(char *app_name)
{
	srand(time(NULL));
	stuff::app_name = new char[strlen(app_name) + 1];
	strcpy(stuff::app_name, app_name);
	file_list = NULL; files_num = 0;
	curr_file = curr_path = path_and_file = NULL;
	help_window_active = FALSE;
	
	if(!read_dir())
	{
		print_error("Error while reading files.");
		exit(-1);
	}
}

stuff::~stuff(void)
{
	stuff::clean_lists();
}

stuff::stuff(void)
{
	stuff::file_list = NULL;
	stuff::files_num = 0;
}


void stuff::clean_lists(void)
{
	int	i;

	if(file_list)
	{
		for(i = 0; i < files_num; i++)
		{
			if(file_list[i])
			{
				delete file_list[i];
				file_list[i] = NULL;
			}
		}
		delete file_list;
		file_list = NULL;
	}
	
	stuff::files_num = 0;
}

bool stuff::read_dir(void)
{
	int		idx_counter = 0, files_num, *idx_list, i;
	struct dirent	**namelist;

	files_num = scandir(".", &namelist, 0, alphasort);
	if(files_num <  0)
		return FALSE;

	idx_list = new int[files_num];
	
	for(i = 0; i < files_num; i++)
	{
		if(is_dir(namelist[i]->d_name) || is_mp3(namelist[i]->d_name) || is_plist(namelist[i]->d_name))
		{
			idx_list[idx_counter] = i;
			idx_counter++;
		}
	}

	if(idx_counter > 0)
		stuff::file_list = new char*[idx_counter];
	for(i = 0; i < idx_counter; i++)
	{
		stuff::file_list[i] = new char[strlen(namelist[idx_list[i]]->d_name) + 1];
		strcpy(stuff::file_list[i], namelist[idx_list[i]]->d_name);
	}

	stuff::files_num = idx_counter;
	
	delete idx_list;
	idx_list = NULL;
	
	return TRUE;
}

void stuff::print_error(char *error_msg)
{
	
	cerr << stuff::app_name << ": " << error_msg << endl;
}

bool stuff::is_dir(char *filename)
{
	struct stat stat_buf;

	stat(filename, &stat_buf);
	if(S_ISDIR(stat_buf.st_mode))
		return TRUE;
	else
		return FALSE;
}

bool stuff::is_link(char *filename)
{
	struct stat stat_buf;

	lstat( filename, &stat_buf );

	if( S_ISLNK(stat_buf.st_mode) )
		return true;
	else
		return false;
}

bool stuff::is_mp3(char *filename)
{
	int	len, i;
	char	*temp;

	temp = new char[5];

	len = strlen(filename);

	for(i = 0; i < 4; i++)
	{
		temp[i] = filename[len - 4 + i];
	}

	temp[4] = (char)NULL;

	if((!strcmp(temp, ".mp3")) || (!strcmp(temp, ".MP3")))
	{
		delete temp;
		return TRUE;
	}
	else
	{
		delete temp;
		return FALSE;
	}
}

int stuff::rfiles_num(void)
{
	return files_num;
}

char *stuff::rfile_name(int indx)
{
	return file_list[indx];
}

void stuff::event_treat(int key, int status, XA_Message msg, class screen &scr, class player &pl, class playlist &plist)
{
	if(key > 0)
	{
		switch(key)
		{
			case	KEY_DOWN:
			case	'j':
			case	'J':
			{
				scr.key_down(this, plist, pl);
				break;
			}
			case	KEY_UP:
			case	'k':
			case	'K':
			{
				scr.key_up(this, plist, pl);
				break;
			}
			case	KEY_NPAGE:
			{
				scr.key_npage(this, plist, pl);
				break;
			}
			case	KEY_PPAGE:
			{
				scr.key_ppage(this, plist, pl);
				break;
			}
			case	KEY_END:
			{
				scr.key_end(this, plist, pl);
				break;
			}
			case	KEY_HOME:
			{
				scr.key_home(this, plist, pl);
				break;
			}
			case	KEY_RIGHT:
			{
				pl.seek(FORWARD, scr);
				break;
			}
			case	KEY_LEFT:
			{
				pl.seek(BACKWARD, scr);
				break;
			}
			case	'.':
			{
				pl.set_volume(FORWARD, scr);
				scr.draw_screen(*this, plist, pl);
				break;
			}
			case	',':
			{
				pl.set_volume(BACKWARD, scr);
				scr.draw_screen(*this, plist, pl);
				break;
			}
			case	'T':
			case	't':
			{
				int	temp;
				char	*curr_file;
				
				curr_file = this->file_list[scr.ridx() + scr.rcursor()];
				if(is_dir(curr_file))
				{
					scr.key_down(this, plist, pl);
					break;
				}
				if(is_plist(curr_file))
				{
					scr.key_down(this, plist, pl);
					break;
				}
				temp = plist.in_playlist(curr_file);

				if(temp < 0)
					plist.add_to_playlist(curr_file);
				else
					plist.remove_from_playlist(temp);
				scr.key_down(this, plist, pl);
				break;
			}
			case	'R':
			case	'r':
			{
				scr.clr_refresh(this, plist, pl);
				break;
			}
			case	RET_KEY:
			{
				if(curr_file)
					delete curr_file;
				curr_file = new char[strlen(this->file_list[scr.ridx() + scr.rcursor()]) + 1];
				strcpy(curr_file, this->file_list[scr.ridx() + scr.rcursor()]);
				
				if(is_dir(curr_file))
				{
					bool	b;

					b = !strcmp( curr_file, ".." );
					
					if( b == false )
						dir_stack.push( strdup(curr_file) );
					
					if( is_link(curr_file) )
					{
						chdir( curr_file );
						
						while( !dir_stack.empty() )
							dir_stack.pop();
					}
					else
						chdir( curr_file );
					
					clean_lists();
					read_dir();
					
					scr.set_params(0, 0);
					scr.key_down_without_draw(this, plist, pl);

					if ( b == true && !dir_stack.empty() )
					{
						char	*t;
						int	i = 0;
						bool	found = false;
						
						t = dir_stack.top(); 
						dir_stack.pop();
						
						while( !found && i < files_num )
						{
							if ( !strcmp(file_list[i], t) )
								found = true;
							i++;
						}

						i--;

						if ( found == true ) 
						{
							scr.set_params(0, 0);
					
							for( int p = 0; p < i; p++ )
								scr.key_down_without_draw(this, plist, pl);
						}
						
						free( t ); //strdup uses malloc
					}
					else
					{
						scr.set_params(0, 0);
						scr.key_down_without_draw( this, plist, pl );
					}
					
					scr.clr_refresh(this, plist, pl);
					break;
				}
				
				if(is_plist(curr_file))
				{
					plist.load_playlist(curr_file);
					scr.draw_screen(*this, plist, pl);
					break;
				}
				
				if(plist.is_playing() && pl.is_paused())
					pl.play("Baaa...");
				else
					if(plist.is_playing())
						return;
				give_path_and_file(curr_file);
				pl.play(path_and_file);
				
				break;
			}
			case	'S':
			case	's':
			{
				if(plist.is_playing())
					plist.reset_playing();
				pl.stop();
				break;
			}
			case	'P':
			case	'p':
			{
				pl.pause();
				break;
			}
			case	'L':
			case	'l':
			{
				if(pl.is_idle())
					plist.play_the_list(pl, FALSE);
				else
					if(pl.is_paused())
						pl.play("Fhrrr...");
				break;
			}
			case	'A':
			case	'a':
			{
				if(!plist.is_playing())
				{
					plist.toggle_play_method();
					scr.draw_screen(*this, plist, pl);
				}
				break;
			}
			case	'M':
			case	'm':
			{
				if(!plist.is_playing())
				{
					add_all_to_playlist(plist);
					scr.draw_screen(*this, plist, pl);
				}
				break;
			}
			case	'U':
			case	'u':
			{
				if(!plist.is_playing())
				{
					remove_all_in_cur_dir(plist);
					scr.draw_screen(*this, plist, pl);
				}
				break;
			}
			case	'C':
			case	'c':
			{
				if(!plist.is_playing())
				{
					plist.remove_all_from_playlist();
					scr.draw_screen(*this, plist, pl);
				}
				break;
			}
			case	'N':
			case	'n':
			{
				if(plist.is_playing())
				{
					if(plist.rplay_method() == RANDOM)
					{
						pl.stop();
						plist.play_the_list(pl, FALSE);
					}
					else
					{
						pl.stop();
						plist.play_the_list(pl, TRUE);
					}
				}
				break;
			}
			case	'B':
			case	'b':
			{
				if(plist.is_playing() && plist.rplay_method() != RANDOM)
				{
					pl.stop();
					plist.play_prev_song_in_list(pl, TRUE);
				}
				break;
			}
			case	'[':
			{
				if(plist.is_playing() && plist.rplay_method() != RANDOM)
				{
					pl.stop();
					plist.goto_first_song_in_list(pl, TRUE);
				}
				break;
			}
			case	']':
			{
				if(plist.is_playing() && plist.rplay_method() != RANDOM)
				{
					pl.stop();
					plist.goto_last_song_in_list(pl, TRUE);
				}
				break;
			}
			case	'W':
			case	'w':
			{
				plist.save_playlist(scr, *this, pl);
				break;
			}
			case	'H':
			case	'h':
			{
				help_screen(scr, pl, plist);
				break;
			}
		}
	}
	else
	{	
		if(status < 0)
		{
			if(status == XA_ERROR_PIPE_CLOSED)
			{
				print_error("Decoder process killed. exiting...");
				exit(-1);
			}
			else
			{
				char	*buffer;
				
				buffer = new char[4096];
				sprintf(buffer, "Error %d gzzz...", status);
				print_error(buffer);
				exit(-1);
			}
		}
		else
		{
			switch(msg.code)
			{
				case	XA_MSG_NOTIFY_READY:
				{
					pl.set_state(IDLE);
					break;
				}
				case	XA_MSG_NOTIFY_EXITED:
				{
					pl.free_decoder();
					exit(-1);
					break;
				}
				case	XA_MSG_NOTIFY_PLAYER_STATE:
				{
					switch(msg.data.state)
					{
						case	XA_PLAYER_STATE_PLAYING:
						{
							pl.set_state(PLAYING);
							scr.draw_screen(*this, plist, pl);
							break;
						}		
						case	XA_PLAYER_STATE_PAUSED:
						{
							pl.set_state(PAUSE);
							scr.draw_screen(*this, plist, pl);
							break;
						}
						case	XA_PLAYER_STATE_STOPPED:
						{
							pl.set_state(IDLE);
							if(!plist.is_playing())
								scr.draw_screen(*this, plist, pl);
							break;
						}
						case	XA_PLAYER_STATE_EOF:
						{
							pl.set_state(IDLE);
							plist.play_the_list(pl, TRUE);
							if(!plist.is_playing())
								scr.draw_screen(*this, plist, pl);
							break;
						}
					}
					break;
				}
				case	XA_MSG_NOTIFY_ACK:
				{
					break;
				}
				case	XA_MSG_NOTIFY_INPUT_POSITION:
				{
					scr.set_song_range_and_offset(msg.data.position.range, msg.data.position.offset);
					scr.draw_screen(*this, plist, pl);
					break;
				}
				case	XA_MSG_NOTIFY_INPUT_TIMECODE:
				{
					char	code[25];

					sprintf(code, "[%.2d:%.2d:%.2d]", msg.data.timecode.h,
							msg.data.timecode.m,
							msg.data.timecode.s);
					
					scr.set_song_timecode(code);
					scr.draw_screen(*this, plist, pl);
					break;
				}
				case	XA_MSG_NOTIFY_INPUT_DURATION:	
				{
					int	hour, minute, second;
					char	code[25];

					hour = msg.data.duration / 3600;
					minute = (msg.data.duration - (hour * 3600)) / 60;
					second = msg.data.duration % 60;
					sprintf(code, "[%.2d:%.2d:%.2d]", hour, minute, second);
					
					scr.set_song_duration(code);
					scr.draw_screen(*this, plist, pl);
					break;
				}
				case	XA_MSG_NOTIFY_INPUT_STREAM_INFO:
				{
					char	info[256];

					sprintf(info, "[MPEG %d, layer %d, %d kbps, %d Hz]",
							msg.data.stream_info.level,
							msg.data.stream_info.layer,
							msg.data.stream_info.bitrate,
							msg.data.stream_info.frequency);
					scr.set_song_info(info);
					break;
				}
				case	XA_MSG_NOTIFY_OUTPUT_CAPS:
				{
					break;
				}
				case	XA_MSG_NOTIFY_OUTPUT_BALANCE:
				{
					scr.set_balance(msg.data.volume.balance);
					break;
				}
				case	XA_MSG_NOTIFY_OUTPUT_PCM_LEVEL:
				{
					scr.set_volume(msg.data.volume.pcm_level);
					break;
				}
				case	XA_MSG_NOTIFY_OUTPUT_MASTER_LEVEL:
				{
					scr.set_master_volume(msg.data.volume.master_level);
					break;
				}
				case	XA_MSG_NOTIFY_ERROR:
				{
					char	err_msg[512];

					sprintf(err_msg," gzz.. error %d (%s)",
							msg.data.error.code,
							msg.data.error.message);
					print_error(err_msg);
					break;
				}
				case	XA_MSG_NOTIFY_DEBUG:
				{
					char	dbg_msg[512];

					sprintf(dbg_msg, "debug message (src=%d,level=%d): %s",
							msg.data.debug.source,
							msg.data.debug.level,
							msg.data.debug.message);
					print_error(dbg_msg);
					break;
				}
			}
		}
	}
}

void stuff::give_path_and_file(char *curr_file)
{
	if(curr_path)
		delete curr_path;
	if(path_and_file)
		delete path_and_file;
	
	curr_path = get_current_dir_name();
	path_and_file = new char [strlen(curr_path) + strlen(curr_file) + 2];
	strcpy(path_and_file, curr_path);
	strcat(path_and_file, "/");
	strcat(path_and_file, curr_file);
}

void stuff::add_all_to_playlist(class playlist &plist)
{
	int	i;

	
	for(i = 0; i < files_num; i++)
	{
		if((!is_dir(file_list[i])) && (!is_plist(file_list[i])))
			plist.add_to_playlist(file_list[i]);
	}
}

void stuff::remove_all_in_cur_dir(playlist &plist)
{
	int	i, i2;

	
	for(i = 0; i < files_num; i++)
	{
		if((!is_dir(file_list[i])) && (!is_plist(file_list[i])))
		{
			i2 = plist.in_playlist(file_list[i]);
			if(i2 != -1)
				plist.remove_from_playlist(i2);
		}
	}
}

bool stuff::is_plist(char *filename)
{
	int	len, i;
	char	*temp;

	temp = new char[5];

	len = strlen(filename);

	for(i = 0; i < 4; i++)
	{
		temp[i] = filename[len - 4 + i];
	}

	temp[4] = (char)NULL;

	if((!strcmp(temp, ".npl")))
	{
		delete temp;
		return TRUE;
	}
	else
	{
		delete temp;
		return FALSE;
	}
}

bool stuff::is_help_window_active(void)
{
	return help_window_active;
}

void stuff::help_screen(screen &scr, player &pl, playlist &plist)
{
	WINDOW	*help_window;

	help_window = scr.rhelp_window();

	wattrset(help_window, A_BOLD|COLOR_PAIR(1));
	box(help_window, 0, 0);

	wattrset(help_window, COLOR_PAIR(6));
	mvwaddstr(help_window, 1, 1, "newplayer, version 2.0");
	wattrset(help_window, A_BOLD|COLOR_PAIR(6));
	mvwhline(help_window, 2, 1, ACS_HLINE, strlen("newplayer, version 2.0"));
	wattrset(help_window, COLOR_PAIR(3));
	mvwaddstr(help_window, 4, 1, "Enter - Play file/Load playlist");
	mvwaddstr(help_window, 5, 1, "L/l   - Play playlist");
	mvwaddstr(help_window, 6, 1, "T/t   - Tag files to playlist");
	mvwaddstr(help_window, 7, 1, "P/p   - Pause");
	mvwaddstr(help_window, 8, 1, "S/s   - Stop");
	mvwaddstr(help_window, 9, 1, "W/w   - Save playlist");
	mvwaddstr(help_window, 10, 1, "A/a   - Toggle play method");
	mvwaddstr(help_window, 11, 1, "C/c   - Clear playlist");
	mvwaddstr(help_window, 12, 1, "R/r   - Refresh screen");
	mvwaddstr(help_window, 13, 1, "B/b   - Previous song in playlist");
	mvwaddstr(help_window, 14, 1, "N/n   - Next song in playlist");
	mvwaddstr(help_window, 15, 1, "M/m   - Tag all files to playlist");
	mvwaddstr(help_window, 16, 1, "U/u   - Untag all in current directory");
	mvwaddstr(help_window, 17, 1, "[     - First song in playlist");
	mvwaddstr(help_window, 18, 1, "]     - Last song in playlist");
	mvwaddstr(help_window, 19, 1, ".     - Increase volume");
	mvwaddstr(help_window, 20, 1, ",     - Decrease volume");
	mvwaddstr(help_window, 21, 1, "Left/Right keys to seek in the song");
	mvwaddstr(help_window, 22, 1, "Browse:");
	mvwaddstr(help_window, 22, 1, "  j/k, Up/Down, PgUp/PgDn, Home/End");
	
	help_window_active = TRUE;
	timeout(-1);
	scr.draw_screen(*this, plist, pl);
	getch();
	scr.draw_screen(*this, plist, pl);
	timeout(0);
	help_window_active = FALSE;
}

void stuff::set_app_name(char *app)
{
	app_name = new char[strlen(app) + 1];
	strcpy(app_name, app);
}
